;------------------------------------------
;Funkcje obsugi pamici 24C64
;------------------------------------------
;Dla mikrokontrolera z rodziny Intel 8051
;pracujcego z zegarem 12MHz

$INCLUDE (REG_51.PDF)			;doczenie definicji rejestrw uK (kompilator RA-51 Raisonance)

NAME	OBSLUGA_AT24C64

FIXEDADDR	EQU	0A0H			;staa czc adresu pamici 24C64
USRADDR		EQU	0				;ustawiana przez uytkownika cz adresu (wartoci 0..7)
MEMSIZE		EQU	2000H			;liczba bajtw dla 24C64
PMEMSIZE		EQU	32				;liczba bajtw na stron pamici dla 24C64
FILL			EQU	0AAH			;warto uywana w funkcji FILL do wypenienia pamici

;Definicje funkcji rejestrw
index			EQU	R0				;wskanik bufora w pamici RAM mikrokontrolera
count			EQU	R1				;licznik bajtw
zdata			EQU	R2				;rejestr bajtu danych
addr_lo		EQU	R3				;modszy bajt adresu bajtu w pamici 24C64
addr_hi		EQU	R4				;starszy bajt - / / -    - / / -

;Definicje linii interfejsu
SCL			BIT	P1.2			;linia zegara (SCL)
SDA			BIT	P1.3			;linia danych (SDA)
ERROROUT		BIT	P1.4			;sygnalizacja bdu weryfikacji


;zmienne w pamici RAM
DSEG AT 20H
buffer:		DS	PMEMSIZE			;bufor do zapisu / odczytu danych
DSEG AT 60H							;pocztek stosu
stack:		DS	20H				;rozmiar stosu

;tablica wektorw przerwa
CSEG

CODE AT 0000H						;wektor przerwania po RESET
				ajmp	on_reset
CODE AT 0003H						;zewntrzne przerwanie INT0
				reti
CODE AT 000BH						;przerwanie TF0
				reti
CODE AT 0013H						;zewntrzne przerwanie INT1
				reti
CODE AT 001BH						;przerwanie TF1
				reti
CODE AT 0023H						;przerwanie od UART
				reti

USING	0								;uywany bdzie bank rejestrw numer 0
CODE AT 0080H						;pocztek programu gwnego

on_reset:
		mov	SP,#(stack-1)		;inicjalizacja wskanika stosu
		setb	SDA					;inicjalizacji linii interfejsu
		setb	SCL					;SDA = SCL = 1
		clr	ERROROUT				;ustawienie stanu niskiego wyjcia sygn.bdu
;< przykady uycia funkcji >
		call	byte_fill			;wypenienie pamici wartoci FILL (0AAH)
		jc		fault					;jeli ustawiona flaga C, to bd
		call	verify_byte_fill	;weryfikacja zawartoci
		jc		fault
		call	page_fill			;wypenienie strony pamici wartoci FILL
		jc		fault					;jeli ustawiona flaga C, to bd
		call	verify_page_fill	;weryfikacja zawartoci
		jc		fault
fault:
		setb	ERROROUT				;sygnalizacja bdu
		jmp	$						;ptla nieskoczona

;wypenienie pamici wartoci zadeklarowan jako FILL
;tryb stronnicowania nie jest wykorzystywany, zapisywany
;jest bajt po bajcie; ustawienie flagi CY wskazuje na przekroczenie
;limitu czasu operacji
;modyfikuje wartoci A, B, DPTR, ADDR_HI (R2), ADDR_LO (R3), bit C
byte_fill:
		mov	zdata,#FILL
		mov	dptr,#0				;adres pocztkowy dla operacji zapisu
bf_x1:
		mov	addr_lo,DPL			;ustawienie adresu bajtu w pamici 24C64
		mov	addr_hi,DPH
		mov	B,#120				;licznik powtrze
bf_x2:
		mov	A,#USRADDR			;do akumulatora zapisywana jest zmienna cz adresu
		call	write_byte			;prba zapisu
		jnc	bf_x3					;skok, jeli zapis pomylny

		djnz	B,bf_x2				;bd,ponowna prba zapisu
		setb	C						;zapis nie powid si, sygnalizacja bdu
		jmp	bf_x4					;i wyjcie
bf_x3:
		inc	dptr					;nastpna pozycja w pamici 24C64
		mov	A,DPL					;sprawdzenie modszego bajtu (limitu) dla adresw
		cjne	A,#(LOW MEMSIZE),bf_x1	;powtrnie, jeli to nie ostatni bajt
		mov	A,DPH					;sprawdzenie starszego bajtu adresu
		cjne	A,#(HIGH MEMSIZE),bf_x1	;powtrnie, jeli to nie ostatni bajt
		clr	C						;zerowanie flagi sygnalizacji bdu
bf_x4:
		ret							;koniec, pomylna realizacja funkcji

;weryfikacja bajtw zapisanych w pamici; odczytuje i sprawdza jeden
;bajt w danym momencie; uywana jest funkcja Random Read do sprawdzenia
;zawartoci pod pierwszym adresiem oraz inicjalizacji wewntrznego licznika
;adresw; nastpnie wywoywana jest Current Address Read i sprawdzana krok
;po kroku zawarto pamici
;modyfikuje wartoci A, B, DPTR, ADDR_HI (R2), ADDR_LO (R3), bit C
verify_byte_fill:
		mov	DPTR,#0				;inicjalizacja rejestru adresw
		mov	addr_lo,DPL			;zapamitanie adresw
		mov	addr_hi,DPH
		mov	B,#120				;liczba powtrze
vbf_x1:
		mov	A,#USRADDR			;zaadowanie zmiennej czci adresu
		call	read_random			;prba odczytu
		jnc	vbf_x2				;skok, jesli pomylna

		djnz	B,vbf_x1				;bd, powtrna prba odczytu
		jmp	vbf_x6				;przekroczono limit, wyjcie z ustawion flag CY
vbf_x2:
		cjne	A,#FILL,vbf_x6		;skok, jeli bd porwnania
		jmp	vbf_x5				;nastpny adres w pamici
vbf_x3:
		mov	A, #USRADDR
		call	read_current		;wywoanie funkcji odczytu biecego bajtu
		jc		vbf_x6				;skok, jeli operacja nie powioda si

		cjne	A,#FILL,vbf_x6		;skok, jeli bd porwnania
vbf_x5:
		inc	dptr					;nastpny adres w pamici 
		mov	A,DPL					;sprawdzenie modszego bajtu (limitu) dla adresw
		cjne	A,#(LOW MEMSIZE),vbf_x3		;powtrnie, jeli to nie ostatni bajt
		mov	A,DPH					;sprawdzenie starszego bajtu dla adresw
		cjne	A,#(HIGH MEMSIZE),vbf_x3	;powtrnie, jeli to nie ostatni bajt
		clr	C						;zerowanie flagi sygnalizacji bdu
		ret							;wyjcie
vbf_x6:
		setb	C						;ustawienie flagi sygnalizacji bdu
		ret

;wypenienie strony w pamici wartoci FILL
;modyfikuje wartoci A, B, DPTR, ADDR_HI (R2), ADDR_LO (R3), bit C
page_fill:
		mov	B,#PMEMSIZE			;do B liczba bajtw na stron
		mov	index,#buffer		;"index" jako wskanik do bufora w RAM mikrokontrolera
pf_x1:
		mov	@index,#FILL		;wypenienie bufora w RAM wartoci FILL
		inc	index
		djnz	B,pf_x1
										;zapis bufora do pamici 24C64 - jedna strona w danym momencie
		mov	dptr,#0				;inicjalizacji wskanika adresw
pf_x2:
		mov	addr_lo,DPL			;nastawy rejestru adresw
		mov	addr_hi,DPH
		mov	count,#PMEMSIZE	;liczba bajtw na stron
		mov	B,#120				;rejestr B jako licznik powtrze
pf_x3:
		mov	A,#USRADDR			;zaadowanie programowanej czci adresu
		call	write_block			;prba zapisu
		jnc	pf_x4					;skok, jeli prba pomylna
		djnz	B,pf_x3				;ponowna prba zapisu
		setb	C						;ustawienie flagi sygnalizacji bdu
		jmp	pf_x6					;i wyjcie
pf_x4:
										;dodanie rozmiaru PMEMSIZE do wskanika adresw
		mov	A,DPL					;dodawanie 2 bajtw (modszy, pniej starszy)
		add	A,#PMEMSIZE
		mov	DPL,A
		jnc	pf_x5
		inc	DPH					;rozkaz INC DPH zastpuje dodanie do DPH przeniesienia
pf_x5:
		cjne	A,#(LOW MEMSIZE),pf_x2	;jak poprzednio-sprawdzenie adresu
		mov	A,DPH
		cjne	A,#(HIGH MEMSIZE),pf_x2
		clr	C
pf_x6:
		ret

;weryfikacja strony zapisanej w pamici 24C64
;weryfikowana jest jedna strona w danym momencie
;modyfikuje wartoci A, B, DPTR, ADDR_HI (R2), ADDR_LO (R3), bit C
verify_page_fill:
						;przepisanie strony z pamici 24C64 do bufora
		mov	DPTR,#0				;inicjalizacja licznika adresw
vpf_x1:
		mov	addr_lo,DPL
		mov	addr_hi,DPH
		mov	count,#PMEMSIZE	;liczba bajtw na stron
		mov	B,#120				;licznik powtrze
vpf_x2:
		mov	A,#USRADDR			;zapamitanie zmiennej czci adresu
		call	read_block			;prba odczytu
		jnc	vpf_x4				;skok, jeli pomylna
		djnz	B,vpf_x2				;ponowna prba odczytu
vpf_x3:
		setb	C						;ustawienie bitu sygnalizacji bdu
		jmp	vpf_x7				;wyjcie
vpf_x4:
										;weryfikacja zawartoci bufora
		mov	B,#PMEMSIZE			;liczba bajtw na stron
		mov	index,#buffer		;zapamietanie adresu bufora w RAM
vpf_x5:
		cjne	@index,#FILL,vpf_x3		;skok, jeli bd porwnania
		inc	index					;nastpna pozycja w buforze
		djnz	B,vpf_x5
										;dodanie PMEMSIZE do wskanika adresu
		mov	A,DPL					;operacja przeprowadzana jak wyej
		add	A,#PMEMSIZE
		mov	DPL,A
		jnc	vpf_x6
		inc	DPH
vpf_x6:
		cjne	A,#(LOW MEMSIZE),vpf_x1
		mov	A,DPH
		cjne	A,#(HIGH MEMSIZE),vpf_x1
		clr	C
vpf_x7:
		ret

;zapis na stronie pamici zawartoci bufora BUFFER ;wywoanie:
;adres 1-go bajtu w ADDR_HI:ADDR_LO, dane do zapisu w BUFFER,
;liczba bajtw w COUNT, zmienna cz adresu ukadu w A
;zwraca CY=1, jeli magistrala jest zajta lub pami nie odpowiada
;modyfikuje: A, COUNT, INDEX
write_block:
		call	start					;wysanie polecenie I2C START
		jc		wb_x8					;przerwa, jeli magistrala jest zajta
		rl		A						;zmienna cz adresu przesuwana jest na pozycje bitw 3:1
		orl	A,#FIXEDADDR		;dodanie staej czci adresu
		clr	ACC.0					;zerowanie bitu numer 0: zapis danych
		call	shout					;wysanie adresu pamici
		jc		wb_x7					;jeli ukad nie odpowiada, to bd i wyjcie
		mov	A,addr_hi			;wysanie starszego bajtu adresu sowa w pamici
		call	shout
		jc		wb_x7					;jeli ukad nie odpowiada, to bd i wyjcie
		mov	A,addr_lo			;wysanie modszego bajtu adresu sowa w pamici
		call	shout
		jc		wb_x7					;jeli ukad nie odpowiada, to bd i wyjcie
		mov	index,#buffer		;zaadowanie do zmiennej INDEX adresu bufora w RAM
wb_x6:
		mov	A,@index				;pobierz bajt
		call	shout					;wylij do pamici 24C64
		jc		wb_x7					;jeli ukad nie odpowiada, to bd i wyjcie
		inc	index					;nastpna pozycja w buforze
		djnz	count,wb_x6			;nastpny adres w 24C64
		clr	C						;zerowanie flagi C - sygnalizacji bdu
wb_x7:
		call	stop					;wysanie polecenie I2C STOP
wb_x8:
		ret

;odczyt jednego bajtu do jednej strony bajtw z pamici 24C64
;wysya polecenie Random Read a nastpnie  Sequential Read
;wywoanie: adres 1-go bajtu w ADDR_HI:ADDR_LO, dane zwracane w BUFFER,
;liczba bajtw w COUNT, zmienna cz adresu ukadu w A
;zwraca CY=1, jeli magistrala jest zajta lub pami nie odpowiada
;modyfikuje: A, COUNT, INDEX
read_block:
										;tryb zapisu w celu ustawienia licznika adresw 24C64
		call	start					;wysanie polecenia I2C START
		jc		rb_x5					;przerwa, jeli magistrala jest zajta
		rl		A						;zmienna cz adresu przesuwana jest na pozycje bitw 3:1
		orl	A,#FIXEDADDR		;dodanie staej czci adresu
		mov	index,A				;zapamitanie kopii adresu urzdzenia
		clr	ACC.0					;wybr operacji zapisu
		call	shout					;wysanie adresu pamici
		jc		rb_x4					;wyjcie, jeli brak potwierdzenia
		mov	A,addr_hi			;wysanie starszej czci adresu sowa
		call	shout
		jc		rb_x4					;wyjcie, jeli brak potwierdzenia
		mov	A,addr_lo			;wysanie modszej czci adresu sowa
		call	shout
		jc		rb_x4					;wyjcie, jeli brak potwierdzenia
										;zmiana trybu na "odczyt" i odczyt bajtw
		call	start					;powtrne wysanie polecenia I2C START
		jc		rb_x4					;wyjcie, jeli brak potwierdzenia
		mov	A,index				;odtworzenie adresu pamici
		setb	ACC.0					;kasowanie bitu 0 adresu - odczyt
		call	shout					;wysanie adresu na magistral I2C
		jc		rb_x4					;wyjcie, jeli brak potwierdzenia
		mov	index,#buffer		;do zmiennej INDEX wskazanie do bufora
rb_x1:
		call	shin					;odbir bajtu od pamici
		mov	@index,A				;zapamitanie ich w RAM mikrokontrolera
		cjne	count,#1,rb_x2		;skok, jeli to nie ostatni bajt
		call	NAK					;nie wysyaj potwierdzenia, ostatni bajt
		jmp	rb_x3					;wyjcie
rb_x2:
		call	ACK					;wylij potwierdzenie
		inc	index					;nastpna pozycja w buforze
		djnz	count,rb_x1			;nastpny bajt w pamici 24C64
rb_x3:
		clr	C						; clear error flag
rb_x4:
		call	stop					;wysanie rozkazu I2C STOP
rb_x5:
		ret

;zapis bajtu do pamici 24C64; wywoanie: adres 1-go bajtu w 
;ADDR_HI:ADDR_LO, dane do zapisu w ZDATA, zmienna cz adresu 
;ukadu pamici w A, zwraca CY=1, jeli magistrala jest zajta lub 
;pami nie odpowiada, modyfikuje A
write_byte:
		call	start					;wysanie rozkazu I2C START
		jc		wb_y9					;wyjcie, jeli brak potwierdzenia
		rl		A						;zmienna cz adresu przesuwana jest na pozycje bitw 3:1
		orl	A,#FIXEDADDR		;dodanie staej czci adresu
		clr	ACC.0					;operacja zapisu (b0=0)
		call	shout					;wysanie adresu pamici
		jc		wb_y8					;wyjcie, jeli brak potwierdzenia
		mov	A,addr_hi			;wysanie starszego bajtu adresu sowa
		call	shout	
		jc		wb_y8					;wyjcie, jeli brak potwierdzenia
		mov	A,addr_lo			;wysanie modszego bajtu adresu sowa
		call	shout
		jc		wb_y8					;wyjcie, jeli brak potwierdzenia
		mov	A,zdata				;pobranie danych z ZDATA do ACC
		call	shout					;wysanie danych do pamici 24C64
		jc		wb_y8					;wyjcie, jeli brak potwierdzenia
		clr	C						;kasowanie znacznika bdu
wb_y8:
		call	stop					;wysanie rozkazu I2C STOP
wb_y9:
		ret

;odczyt bajtu spod biecego adresu w pamici 24C64
;wywoywana ze zmienn czci adresu w A, zwraca bajt w A
;zwraca CY=1, jeli magistrala jest zajta lub pami nie 
;odpowiada, modyfikuje A
read_current:
		call	start					;wysanie rozkazu I2C START
		jc		rc_x5					;wyjcie, jeli brak potwierdzenia
		rl		A						;zmienna cz adresu przesuwana jest na pozycje bitw 3:1
		orl	A,#FIXEDADDR		;dodanie staej czci adresu
		setb	ACC.0					;operacja odczytu (b0=1)
		call	shout					;wysanie adresu pamici
		jc		rc_x4					;wyjcie, jeli brak potwierdzenia
		call	shin					;odbir bajtu danych
		call	NAK					;nie wysyaj potwierdzenia, koniec operacji
		clr	C						;kasowanie flagi sygnalizacji bdu
rc_x4:
		call	stop					;wysanie rozkazu I2C STOP
rc_x5:
		ret

;wywyanie funkcji Read Random (odczyt swobodny)
;wywoywana ze zmienn czci adresu w A, adresem bajtu w 
;ADDR_HI:ADDR_LO; bajt zwracany w akumulatorze
;zwraca CY=1, jeli magistrala jest zajta lub pami nie 
;odpowiada
read_random:
		push	B						;zapamitanie stanu rejestru B na stosie
		mov	B,A					;zapamitanie kopii zmiennej czci adresu
										;tryb zapisu w celu ustawienia licznika adresw 24C64
		call	start					;wysanie rozkazu I2C START
		jc		rr_x7					;wyjcie, jeli brak potwierdzenia
		rl		A						;zmienna cz adresu przesuwana jest na pozycje bitw 3:1
		orl	A,#FIXEDADDR		;dodanie staej czci adresu
		clr	ACC.0					;operacja zapisu (b0=0)
		call	shout					;wysanie adresu pamici
		jc		rr_x6					;wyjcie, jeli brak potwierdzenia
		mov	A,addr_hi			;wysanie starszego bajtu adresu sowa
		call	shout
		jc		rr_x6					;wyjcie, jeli brak potwierdzenia
		mov	a,addr_lo			;wysanie modszego bajtu adresu sowa
		call	shout
		jc		rr_x6					;wyjcie, jeli brak potwierdzenia
										;wywoanie funkcji Call Current Address Read
		mov	A,B					;odtwrz zmienn cz adresu
		call	read_current
		jmp	rr_x7					;wyjcie
rr_x6:
		call	stop					;wysanie rozkazu I2C STOP
rr_x7:
		pop	B						;odtworzenie stanu rejestru B
		ret

;wysanie polecenie I2C START (SDA = "1" -> "0" przy SCL = "1")
;powrt z SCL = SDA = 0, zwraca CY = 1, jeli magistrala jest zajta
start:
		setb	SDA
		setb	SCL
										;sprawdzenie zajtoci magistrali I2C
		jnb	SDA,st_x0			;skok, jeli SDA <> "1"
		jnb	SCL,st_x0			;skok, jeli SCL <> "1"
		nop							;czas na ustalenie si linii
		clr	SDA					;zmiana stanu SDA (SDA -> "0" przy SCL = "1")
		nop							;opnienie
		nop
		nop
		clr	SCL					;zmiana stanu SCL (SCL -> "0" przy SDA = "0")
		clr	C						;kasowanie flagi sygnalizacji bdu
		jmp	st_x1
st_x0:
		setb	c						;ustawienie flagi sygnalizacji bdu
st_x1:
		ret

;wysanie rozkazu I2C STOP (SDA = "0" -> "1" przy SCL = "1")
;spodziewane jest SCL = "0", powrr z SDA = SCL = "1"
stop:
		clr	SDA
		nop							;czas na ustalenie SDA
		setb	SCL					;SCL -> "1" przy SDA = "0"
		nop							;czas na ustalenie
		nop
		nop
		setb	SDA					;zmiana SDA -> "1"
		ret

;wyprowadzenie bajtu przez SDA w takt SCL (najbardziej znaczcy
;bit wyprowadzany jest jako pierwszy); wywoywana z bajtem do 
;wysania w A, zwraca CY = 1, jeli magistrala jest zajta
shout:
		push	B						;zapamitanie stanu rejestru B na stosie
		mov	B,#8					;rejestr B jako licznik wysyanych bitw
sho_x2:
		rlc	A						;przesunicie bajtu w lewo i przeniesienie najstarszego
										;bitu dla flagi C
		mov	SDA, c				;wyprowadzenie bitu
		nop							;czas na ustalenie si wyjcia mikrokontrolera
		setb	SCL					;zmiana SCL ___|```
		nop							;czas na ustalenie 
		clr	SCL					;zmiana SCL ```|___
		djnz	B,sho_x2				;nastpny bit
		setb	SDA					;ustawienie SDA = "1" dla odbioru ACK
		nop							;czas na ustalenie si pamici
		setb	SCL					;narastajce zbocze SCL
		nop							;czas na ustalenie 
		mov	C,SDA					;odczyt bitu ACK
		clr	SCL					;zmiana SCL "1" -> "0"
		pop	B						;odtworzenie stanu rejestru B
		ret

;odczyt 8 bitw, kompletacja sowa danych (z postaci szeregowej na rwnoleg)
;jako pierwszy wprowadzany jest najbardziej znaczcy bit, bajt zwracany jest w A
shin:
		setb	SDA					;SDA jak wejcie
		push	B						;zapamitanie stanu rejestru B na stosie
		mov	B,#8					;rejestr B jako licznik wysyanych bitw
shi_x3:
		nop							;czas na ustalenie si linii danych
		setb	SCL					;narastajce zbocze SCL
		nop
		mov	C,SDA					;odczyt linii danych
		rlc	A						;wprowadzenie bitu do akumulatora na najmodsz pozycj
		clr	SCL					;opadajce zbocze SCL
		djnz	B,shi_x3				;nastpny bit
		pop	B						;odtworzenie stanu rejestru B
		ret

;wysanie potwierdzenia (stan niski linii SDA)
ACK:
		clr	SDA					;wysanie ACK
		nop							;czas na ustalenie
		setb	SCL					;narastajce zbocze sygnau SCL
		nop							;czas na ustalenie
		clr	SCL					;opadjce zbocze SCL
		ret


;wysanie "nie-potwierdzenia" 
NAK:
		setb	SDA					;ustawienie NAK
		nop							;czas na ustalenie
		setb	SCL					;narastajce zbocze sygnau SCL
		nop							;czas na ustalenie
		clr	SCL					;opadjce zbocze SCL
		ret

END
